from typing import List, Optional

from fastapi import APIRouter, Depends
from starlette.responses import Response

from watchmen_auth import PrincipalService
from watchmen_indicator_kernel.meta import DerivedObjectiveService, ObjectiveService
from watchmen_indicator_surface.settings import ask_tuple_delete_enabled
from watchmen_indicator_surface.util import trans, trans_readonly
from watchmen_meta.admin import UserService
from watchmen_meta.common import ask_meta_storage, ask_snowflake_generator
from watchmen_model.admin import UserRole
from watchmen_model.common import DerivedObjectiveId, ObjectiveId
from watchmen_model.indicator import DerivedObjective, Objective
from watchmen_rest import get_console_principal, get_super_admin_principal, get_principal_by_jwt, \
	retrieve_authentication_manager
from watchmen_rest.util import raise_400, raise_403, raise_404
from watchmen_utilities import get_current_time_in_seconds, is_blank

router = APIRouter()


def get_derived_objective_service(principal_service: PrincipalService) -> DerivedObjectiveService:
	return DerivedObjectiveService(ask_meta_storage(), ask_snowflake_generator(), principal_service)


def get_objective_service(derived_objective_service: DerivedObjectiveService) -> ObjectiveService:
	return ObjectiveService(
		derived_objective_service.storage, derived_objective_service.snowflakeGenerator,
		derived_objective_service.principalService)


def get_user_service(derived_objective_service: DerivedObjectiveService) -> UserService:
	return UserService(
		derived_objective_service.storage, derived_objective_service.snowflakeGenerator,
		derived_objective_service.principalService)


@router.get(
	'/indicator/derived-objective/connect', tags=[UserRole.CONSOLE, UserRole.ADMIN], response_model=DerivedObjective)
async def connect_as_derived_objective(
		objective_id: Optional[ObjectiveId], name: Optional[str],
		principal_service: PrincipalService = Depends(get_console_principal)
) -> DerivedObjective:
	if is_blank(objective_id):
		raise_400('Objective id is required.')

	derived_objective_service = get_derived_objective_service(principal_service)

	def action() -> DerivedObjective:
		objective_service = get_objective_service(derived_objective_service)
		objective: Optional[Objective] = objective_service.find_by_id(objective_id)
		if objective is None:
			raise_400('Incorrect objective id.')
		if objective.tenantId != principal_service.get_tenant_id():
			raise_403()

		derived_objective = DerivedObjective(name=name, objectiveId=objective_id, definition=objective)
		derived_objective_service.redress_storable_id(derived_objective)
		derived_objective.userId = principal_service.get_user_id()
		derived_objective.tenantId = principal_service.get_tenant_id()
		derived_objective.lastVisitTime = get_current_time_in_seconds()
		# noinspection PyTypeChecker
		derived_objective: DerivedObjective = derived_objective_service.create(derived_objective)
		return derived_objective

	return trans(derived_objective_service, action)


@router.post('/indicator/derived-objective', tags=[UserRole.CONSOLE, UserRole.ADMIN], response_model=DerivedObjective)
async def save_derived_objective(
		derived_objective: DerivedObjective, principal_service: PrincipalService = Depends(get_console_principal)
) -> DerivedObjective:
	derived_objective_service = get_derived_objective_service(principal_service)

	# noinspection DuplicatedCode
	def action(a_derived_objective: DerivedObjective) -> DerivedObjective:
		a_derived_objective.userId = principal_service.get_user_id()
		a_derived_objective.tenantId = principal_service.get_tenant_id()
		a_derived_objective.lastVisitTime = get_current_time_in_seconds()
		if derived_objective_service.is_storable_id_faked(a_derived_objective.derivedObjectiveId):
			derived_objective_service.redress_storable_id(a_derived_objective)
			# noinspection PyTypeChecker
			a_derived_objective: DerivedObjective = derived_objective_service.create(a_derived_objective)
		else:
			# noinspection PyTypeChecker
			existing_derived_objective: Optional[DerivedObjective] = \
				derived_objective_service.find_by_id(a_derived_objective.derivedObjectiveId)
			if existing_derived_objective is not None:
				if existing_derived_objective.tenantId != a_derived_objective.tenantId:
					raise_403()
				if existing_derived_objective.userId != a_derived_objective.userId:
					raise_403()

			# noinspection PyTypeChecker
			a_derived_objective: DerivedObjective = derived_objective_service.update(a_derived_objective)

		return a_derived_objective

	return trans(derived_objective_service, lambda: action(derived_objective))


@router.get(
	'/indicator/derived-objective/list', tags=[UserRole.CONSOLE, UserRole.ADMIN], response_model=List[DerivedObjective])
async def find_my_derived_objectives(
		principal_service: PrincipalService = Depends(get_console_principal)
) -> List[DerivedObjective]:
	derived_objective_service = get_derived_objective_service(principal_service)

	def action() -> List[DerivedObjective]:
		# noinspection PyTypeChecker
		derived_objectives: List[DerivedObjective] = derived_objective_service.find_all_by_user_id(
			principal_service.get_user_id(), principal_service.get_tenant_id())
		return derived_objectives

	return trans_readonly(derived_objective_service, action)


@router.get('/indicator/derived-objective/rename', tags=[UserRole.CONSOLE, UserRole.ADMIN], response_class=Response)
async def update_derived_objective_name_by_id(
		derived_objective_id: Optional[DerivedObjectiveId], name: Optional[str],
		principal_service: PrincipalService = Depends(get_console_principal)
) -> None:
	if is_blank(derived_objective_id):
		raise_400('Derived objective id is required.')

	derived_objective_service = get_derived_objective_service(principal_service)

	# noinspection DuplicatedCode
	def action() -> None:
		existing_one = derived_objective_service.find_tenant_and_user(derived_objective_id)
		if existing_one is None:
			raise_404()
		existing_tenant_id, existing_user_id = existing_one
		if existing_tenant_id != principal_service.get_tenant_id():
			raise_403()
		elif existing_user_id != principal_service.get_user_id():
			raise_403()
		# noinspection PyTypeChecker
		derived_objective_service.update_name(
			derived_objective_id, name, principal_service.get_user_id(), principal_service.get_tenant_id())

	trans(derived_objective_service, action)


@router.get('/indicator/derived-objective/delete', tags=[UserRole.CONSOLE, UserRole.ADMIN], response_class=Response)
async def delete_derived_objective_by_id(
		derived_objective_id: Optional[DerivedObjectiveId],
		principal_service: PrincipalService = Depends(get_console_principal)
) -> None:
	if is_blank(derived_objective_id):
		raise_400('Derived objective id is required.')

	derived_objective_service = get_derived_objective_service(principal_service)

	# noinspection DuplicatedCode
	def action() -> None:
		# noinspection PyTypeChecker
		existing_derived_objective: Optional[DerivedObjective] = \
			derived_objective_service.find_by_id(derived_objective_id)
		if existing_derived_objective is None:
			raise_404()
		if existing_derived_objective.tenantId != principal_service.get_tenant_id():
			raise_403()
		if not principal_service.is_tenant_admin() and existing_derived_objective.userId != principal_service.get_user_id():
			raise_403()
		derived_objective_service.delete(derived_objective_id)

	trans(derived_objective_service, action)


@router.delete('/indicator/derived-objective', tags=[UserRole.SUPER_ADMIN], response_model=DerivedObjective)
async def delete_derived_objective_by_id_by_super_admin(
		derived_objective_id: Optional[DerivedObjectiveId] = None,
		principal_service: PrincipalService = Depends(get_super_admin_principal)
) -> DerivedObjective:
	if not ask_tuple_delete_enabled():
		raise_404('Not Found')

	if is_blank(derived_objective_id):
		raise_400('Derived objective id is required.')

	derived_objective_service = get_derived_objective_service(principal_service)

	def action() -> DerivedObjective:
		# noinspection PyTypeChecker
		existing_derived_objective: Optional[DerivedObjective] = \
			derived_objective_service.find_by_id(derived_objective_id)
		if existing_derived_objective is None:
			raise_404()
		derived_objective_service.delete(derived_objective_id)
		return existing_derived_objective

	return trans(derived_objective_service, action)


@router.get('/indicator/derived-objective/shared', tags=[UserRole.CONSOLE, UserRole.ADMIN], response_model=DerivedObjective)
def load_share_derived_objective_by_id_and_token(derived_objective_id:DerivedObjectiveId,token:str)->DerivedObjective:

	principal_service: PrincipalService = get_principal_by_jwt(
		retrieve_authentication_manager(), token, [UserRole.CONSOLE, UserRole.ADMIN])

	derived_objective_service = get_derived_objective_service(principal_service)
	def action() -> DerivedObjective:

		existing_derived_objective: Optional[DerivedObjective] = \
			derived_objective_service.find_by_id(derived_objective_id)
		if existing_derived_objective is None:
			raise_404()
		return existing_derived_objective

	return trans_readonly(derived_objective_service, action)


from typing import Dict, List, Optional

from fastapi import APIRouter, Body, Depends

from watchmen_auth import PrincipalService
from watchmen_indicator_kernel.meta import ConvergenceService
from watchmen_indicator_surface.settings import ask_tuple_delete_enabled
from watchmen_indicator_surface.util import trans, trans_readonly
from watchmen_meta.admin import UserGroupService, UserService
from watchmen_meta.common import ask_meta_storage, ask_snowflake_generator
from watchmen_model.admin import User, UserGroup, UserRole
from watchmen_model.common import ConvergenceId, DataPage, Pageable, TenantId, UserGroupId
from watchmen_model.indicator import Convergence
from watchmen_rest import get_admin_principal, get_console_principal, get_super_admin_principal
from watchmen_rest.util import raise_400, raise_403, raise_404, validate_tenant_id
from watchmen_utilities import ArrayHelper, is_blank, is_not_blank

router = APIRouter()


def get_convergence_service(principal_service: PrincipalService) -> ConvergenceService:
	return ConvergenceService(ask_meta_storage(), ask_snowflake_generator(), principal_service)


def get_user_service(convergence_service: ConvergenceService) -> UserService:
	return UserService(
		convergence_service.storage, convergence_service.snowflakeGenerator, convergence_service.principalService)


def get_user_group_service(convergence_service: ConvergenceService) -> UserGroupService:
	return UserGroupService(
		convergence_service.storage, convergence_service.snowflakeGenerator, convergence_service.principalService)


@router.get('/indicator/convergence/name', tags=[UserRole.CONSOLE, UserRole.ADMIN], response_model=List[Convergence])
async def find_convergences_by_name(
		query_name: Optional[str], principal_service: PrincipalService = Depends(get_console_principal)
) -> List[Convergence]:
	convergence_service = get_convergence_service(principal_service)

	# noinspection DuplicatedCode
	def action() -> List[Convergence]:
		tenant_id: TenantId = principal_service.get_tenant_id()
		convergences: List[Convergence]
		if is_blank(query_name):
			# noinspection PyTypeChecker
			convergences = convergence_service.find_by_text(None, tenant_id)
		else:
			# noinspection PyTypeChecker
			convergences = convergence_service.find_by_text(query_name, tenant_id)

		if principal_service.is_admin():
			# current is admin, returns all
			return convergences
		else:
			# current is not admin, only granted will be returned
			# noinspection PyTypeChecker
			user: User = get_user_service(convergence_service).find_by_id(principal_service.get_user_id())
			group_id_map: Dict[UserGroupId, bool]
			if user.groupIds is not None:
				group_id_map = ArrayHelper(user.groupIds).filter(is_not_blank).to_map(lambda x: x, lambda x: True)
			else:
				group_id_map = {}

			def can_access(convergence: Convergence) -> bool:
				user_group_ids = convergence.groupIds
				if user_group_ids is None or len(user_group_ids) == 0:
					return False
				else:
					return ArrayHelper(user_group_ids).some(lambda x: group_id_map[x] is not None)

			return ArrayHelper(convergences).filter(can_access).to_list()

	return trans_readonly(convergence_service, action)


@router.get('/indicator/convergence/list/name', tags=[UserRole.ADMIN], response_model=List[Convergence])
async def find_convergences_by_name(
		query_name: Optional[str], principal_service: PrincipalService = Depends(get_console_principal)
) -> List[Convergence]:
	convergence_service = get_convergence_service(principal_service)

	def action() -> List[Convergence]:
		tenant_id: TenantId = principal_service.get_tenant_id()
		if is_blank(query_name):
			# noinspection PyTypeChecker
			return convergence_service.find_by_name(None, tenant_id)
		else:
			# noinspection PyTypeChecker
			return convergence_service.find_by_name(query_name, tenant_id)

	return trans_readonly(convergence_service, action)


class QueryConvergenceDataPage(DataPage):
	data: List[Convergence]


@router.post('/indicator/convergence/name', tags=[UserRole.ADMIN], response_model=QueryConvergenceDataPage)
async def find_convergences_page_by_name(
		query_name: Optional[str], pageable: Pageable = Body(...),
		principal_service: PrincipalService = Depends(get_admin_principal)) -> QueryConvergenceDataPage:
	convergence_service = get_convergence_service(principal_service)

	def action() -> QueryConvergenceDataPage:
		tenant_id: TenantId = principal_service.get_tenant_id()
		if is_blank(query_name):
			# noinspection PyTypeChecker
			return convergence_service.find_page_by_text(None, tenant_id, pageable)
		else:
			# noinspection PyTypeChecker
			return convergence_service.find_page_by_text(query_name, tenant_id, pageable)

	return trans_readonly(convergence_service, action)


@router.post('/indicator/convergence/ids', tags=[UserRole.ADMIN], response_model=List[Convergence])
async def find_convergences_by_ids(
		convergence_ids: List[ConvergenceId], principal_service: PrincipalService = Depends(get_console_principal)
) -> List[Convergence]:
	if len(convergence_ids) == 0:
		return []

	convergence_service = get_convergence_service(principal_service)

	def action() -> List[Convergence]:
		tenant_id: TenantId = principal_service.get_tenant_id()
		return convergence_service.find_by_ids(convergence_ids, tenant_id)

	return trans_readonly(convergence_service, action)


@router.get('/indicator/convergence', tags=[UserRole.ADMIN], response_model=Convergence)
async def load_convergence_by_id(
		convergence_id: Optional[ConvergenceId], principal_service: PrincipalService = Depends(get_admin_principal)
) -> Convergence:
	if is_blank(convergence_id):
		raise_400('Convergence id is required.')

	convergence_service = get_convergence_service(principal_service)

	def action() -> Convergence:
		# noinspection PyTypeChecker
		convergence: Convergence = convergence_service.find_by_id(convergence_id)
		if convergence is None:
			raise_404()
		# tenant id must match current principal's
		if convergence.tenantId != principal_service.get_tenant_id():
			raise_404()

		return convergence

	return trans_readonly(convergence_service, action)


# noinspection DuplicatedCode
def has_convergence_id(user_group: UserGroup, convergence_id: ConvergenceId) -> bool:
	if user_group.convergenceIds is None:
		return False
	elif len(user_group.convergenceIds) == 0:
		return False
	else:
		return convergence_id in user_group.convergenceIds


def append_convergence_id_to_user_group(user_group: UserGroup, convergence_id: ConvergenceId) -> UserGroup:
	if user_group.convergenceIds is None:
		user_group.convergenceIds = [convergence_id]
	else:
		user_group.convergenceIds.append(convergence_id)
	return user_group


def update_user_group(user_group_service: UserGroupService, user_group: UserGroup) -> None:
	user_group_service.update(user_group)


# noinspection DuplicatedCode
def sync_convergence_to_groups(
		convergence_service: ConvergenceService,
		convergence_id: ConvergenceId, user_group_ids: List[UserGroupId],
		tenant_id: TenantId
) -> None:
	if user_group_ids is None:
		return

	given_count = len(user_group_ids)
	if given_count == 0:
		# do nothing
		return

	user_group_service = get_user_group_service(convergence_service)
	user_groups = user_group_service.find_by_ids(user_group_ids, tenant_id)
	found_count = len(user_groups)
	if given_count != found_count:
		raise_400('User group ids do not match.')

	ArrayHelper(user_groups) \
		.filter(lambda x: not has_convergence_id(x, convergence_id)) \
		.map(lambda x: append_convergence_id_to_user_group(x, convergence_id)) \
		.each(lambda x: update_user_group(user_group_service, x))


def remove_convergence_id_from_user_group(user_group: UserGroup, convergence_id: ConvergenceId) -> UserGroup:
	user_group.convergenceIds = ArrayHelper(user_group.convergenceIds).filter(lambda x: x != convergence_id).to_list()
	return user_group


# noinspection DuplicatedCode
def remove_convergence_from_groups(
		convergence_service: ConvergenceService,
		convergence_id: ConvergenceId, user_group_ids: List[UserGroupId],
		tenant_id: TenantId
) -> None:
	if user_group_ids is None:
		return

	given_count = len(user_group_ids)
	if given_count == 0:
		# do nothing
		return

	user_group_service = get_user_group_service(convergence_service)
	user_groups = user_group_service.find_by_ids(user_group_ids, tenant_id)
	found_count = len(user_groups)
	if given_count != found_count:
		raise_400('User group ids do not match.')

	ArrayHelper(user_groups) \
		.filter(lambda x: has_convergence_id(x, convergence_id)) \
		.map(lambda x: remove_convergence_id_from_user_group(x, convergence_id)) \
		.each(lambda x: update_user_group(user_group_service, x))


@router.post('/indicator/convergence', tags=[UserRole.ADMIN], response_model=Convergence)
async def save_convergence(
		convergence: Convergence, principal_service: PrincipalService = Depends(get_admin_principal)) -> Convergence:
	validate_tenant_id(convergence, principal_service)
	convergence_service = get_convergence_service(principal_service)

	# noinspection DuplicatedCode
	def action(an_convergence: Convergence) -> Convergence:
		if convergence_service.is_storable_id_faked(an_convergence.convergenceId):
			convergence_service.redress_storable_id(an_convergence)
			user_group_ids = ArrayHelper(an_convergence.groupIds).distinct().to_list()
			an_convergence.groupIds = user_group_ids
			# noinspection PyTypeChecker
			an_convergence: Convergence = convergence_service.create(an_convergence)
			# synchronize convergence to user groups
			sync_convergence_to_groups(
				convergence_service, convergence.convergenceId, user_group_ids, convergence.tenantId)
		else:
			# noinspection PyTypeChecker
			existing_convergence: Optional[Convergence] = convergence_service.find_by_id(an_convergence.convergenceId)
			if existing_convergence is not None:
				if existing_convergence.tenantId != an_convergence.tenantId:
					raise_403()

			user_group_ids = ArrayHelper(an_convergence.groupIds).distinct().to_list()
			an_convergence.groupIds = user_group_ids
			# noinspection PyTypeChecker
			an_convergence: Convergence = convergence_service.update(an_convergence)
			# remove convergence from user groups, in case user groups are removed
			removed_user_group_ids = ArrayHelper(existing_convergence.groupIds).difference(user_group_ids).to_list()
			remove_convergence_from_groups(
				convergence_service, an_convergence.convergenceId, removed_user_group_ids, an_convergence.tenantId)
			# synchronize convergence to user groups
			sync_convergence_to_groups(
				convergence_service, an_convergence.convergenceId, user_group_ids, an_convergence.tenantId)

		return an_convergence

	return trans(convergence_service, lambda: action(convergence))


@router.get(
	'/indicator/convergence/available', tags=[UserRole.CONSOLE, UserRole.ADMIN], response_model=List[Convergence])
async def find_available_convergences(
		principal_service: PrincipalService = Depends(get_console_principal)) -> List[Convergence]:
	convergence_service = get_convergence_service(principal_service)

	# noinspection DuplicatedCode
	def action() -> List[Convergence]:
		tenant_id: TenantId = principal_service.get_tenant_id()
		user_id = principal_service.get_user_id()
		# noinspection PyTypeChecker
		user: User = get_user_service(convergence_service).find_by_id(user_id)
		user_group_ids = user.groupIds
		if user_group_ids is None or len(user_group_ids) == 0:
			return []
		user_group_ids = ArrayHelper(user_group_ids).filter(lambda x: is_not_blank(x)).to_list()
		if len(user_group_ids) == 0:
			return []
		user_group_service = get_user_group_service(convergence_service)
		user_groups = user_group_service.find_by_ids(user_group_ids, tenant_id)

		def gather_convergence_ids(distinct_convergence_ids: List[ConvergenceId], user_group: UserGroup) -> List[
			ConvergenceId]:
			given_convergence_ids = user_group.convergenceIds
			if given_convergence_ids is None or len(given_convergence_ids) == 0:
				return distinct_convergence_ids
			given_convergence_ids = ArrayHelper(given_convergence_ids).filter(lambda x: is_not_blank(x)).to_list()
			for convergence_id in given_convergence_ids:
				if convergence_id not in distinct_convergence_ids:
					distinct_convergence_ids.append(convergence_id)
			return distinct_convergence_ids

		convergence_ids = ArrayHelper(user_groups).reduce(gather_convergence_ids, [])
		return convergence_service.find_by_ids(convergence_ids, tenant_id)

	return trans_readonly(convergence_service, action)


@router.delete('/indicator/convergence', tags=[UserRole.SUPER_ADMIN], response_model=Convergence)
async def delete_convergence_by_id_by_super_admin(
		convergence_id: Optional[ConvergenceId] = None,
		principal_service: PrincipalService = Depends(get_super_admin_principal)
) -> Convergence:
	if not ask_tuple_delete_enabled():
		raise_404('Not Found')

	if is_blank(convergence_id):
		raise_400('Convergence id is required.')

	convergence_service = get_convergence_service(principal_service)

	def action() -> Convergence:
		# noinspection PyTypeChecker
		convergence: Convergence = convergence_service.delete(convergence_id)
		if convergence is None:
			raise_404()
		return convergence

	return trans(convergence_service, action)



from typing import List, Optional

from fastapi import APIRouter, Depends
from starlette.responses import Response

from watchmen_auth import PrincipalService
from watchmen_indicator_kernel.meta import DerivedObjectiveReportService, DerivedObjectiveService, \
	ObjectiveReportService
from watchmen_indicator_surface.util import trans, trans_readonly
from watchmen_meta.admin import UserService
from watchmen_meta.common import ask_meta_storage, ask_snowflake_generator
from watchmen_model.admin import UserRole
from watchmen_model.common import ObjectiveId
from watchmen_model.common.tuple_ids import DerivedObjectiveReportId
from watchmen_model.indicator import DerivedObjective
from watchmen_model.indicator.derived_objective_report import DerivedObjectiveReport
from watchmen_model.indicator.objective_report import ObjectiveReport
from watchmen_rest import get_console_principal
from watchmen_rest.util import raise_400, raise_403, raise_404
from watchmen_utilities import get_current_time_in_seconds, is_blank

router = APIRouter()


def get_derived_objective_report_service(principal_service: PrincipalService) -> DerivedObjectiveReportService:
	return DerivedObjectiveReportService(ask_meta_storage(), ask_snowflake_generator(), principal_service)


def get_objective_report_service(principal_service: PrincipalService) -> ObjectiveReportService:
	return ObjectiveReportService(ask_meta_storage(), ask_snowflake_generator(), principal_service)

def get_user_service(derived_objective_service: DerivedObjectiveService) -> UserService:
	return UserService(
		derived_objective_service.storage, derived_objective_service.snowflakeGenerator,
		derived_objective_service.principalService)


@router.get(
	'/indicator/derived-objective/report/connect', tags=[UserRole.CONSOLE, UserRole.ADMIN],
	response_model=DerivedObjectiveReport)
async def connect_as_derived_objective_report(
		objective_report_id: Optional[ObjectiveId], name: Optional[str],
		principal_service: PrincipalService = Depends(get_console_principal)
) -> DerivedObjectiveReport:
	if is_blank(objective_report_id):
		raise_400('objective report id is required.')

	derived_objective_report_service = get_derived_objective_report_service(principal_service)

	def action() -> DerivedObjectiveReport:
		objective_report_service = get_objective_report_service(principal_service)
		objective_report: Optional[ObjectiveReport] = objective_report_service.find_by_id(objective_report_id)
		if objective_report is None:
			raise_400('Incorrect objective report id.')
		if objective_report.tenantId != principal_service.get_tenant_id():
			raise_403()

		derived_objective_report = DerivedObjectiveReport(name=name, objectiveReportId=objective_report_id,
		                                                  definition=objective_report)
		derived_objective_report_service.redress_storable_id(derived_objective_report)
		derived_objective_report.userId = principal_service.get_user_id()
		derived_objective_report.tenantId = principal_service.get_tenant_id()
		derived_objective_report.lastVisitTime = get_current_time_in_seconds()
		# noinspection PyTypeChecker
		derived_objective_report: DerivedObjectiveReport = derived_objective_report_service.create(
			derived_objective_report)
		return derived_objective_report

	return trans(derived_objective_report_service, action)


@router.post('/indicator/derived-objective-report', tags=[UserRole.CONSOLE, UserRole.ADMIN],
             response_model=DerivedObjectiveReport)
async def save_derived_objective_report(
		derived_objective_report: DerivedObjectiveReport,
		principal_service: PrincipalService = Depends(get_console_principal)
) -> DerivedObjectiveReport:
	derived_objective_report_service = get_derived_objective_report_service(principal_service)

	# noinspection DuplicatedCode
	def action(a_derived_objective_report: DerivedObjectiveReport) -> DerivedObjectiveReport:
		a_derived_objective_report.userId = principal_service.get_user_id()
		a_derived_objective_report.tenantId = principal_service.get_tenant_id()
		a_derived_objective_report.lastVisitTime = get_current_time_in_seconds()
		if derived_objective_report_service.is_storable_id_faked(a_derived_objective_report.derivedObjectiveId):
			derived_objective_report_service.redress_storable_id(a_derived_objective_report)
			# noinspection PyTypeChecker
			a_derived_objective_report: DerivedObjectiveReport = derived_objective_report_service.create(
				a_derived_objective_report)
		else:
			# noinspection PyTypeChecker
			existing_derived_objective_report: Optional[DerivedObjectiveReport] = \
				derived_objective_report_service.find_by_id(a_derived_objective_report.derivedObjectiveReportId)
			if existing_derived_objective_report is not None:
				if existing_derived_objective_report.tenantId != a_derived_objective_report.tenantId:
					raise_403()
				if existing_derived_objective_report.userId != a_derived_objective_report.userId:
					raise_403()

			# noinspection PyTypeChecker
			a_derived_objective_report: DerivedObjective = derived_objective_report_service.update(
				a_derived_objective_report)

		return a_derived_objective_report

	return trans(derived_objective_report_service, lambda: action(derived_objective_report))


@router.get(
	'/indicator/derived-objective-report/list', tags=[UserRole.CONSOLE, UserRole.ADMIN],
	response_model=List[DerivedObjectiveReport])
async def find_my_derived_objective_reports(
		principal_service: PrincipalService = Depends(get_console_principal)
) -> List[DerivedObjectiveReport]:
	derived_objective_report_service = get_derived_objective_report_service(principal_service)

	def action() -> List[DerivedObjectiveReport]:
		# noinspection PyTypeChecker
		derived_objectives_reports: List[DerivedObjectiveReport] = derived_objective_report_service.find_all_by_user_id(
			principal_service.get_user_id(), principal_service.get_tenant_id())
		return derived_objectives_reports

	return trans_readonly(derived_objective_report_service, action)


@router.get('/indicator/derived-objective-report/rename', tags=[UserRole.CONSOLE, UserRole.ADMIN],
            response_class=Response)
async def update_derived_objective_report_name_by_id(
		derived_objective_report_id: Optional[DerivedObjectiveReportId], name: Optional[str],
		principal_service: PrincipalService = Depends(get_console_principal)
) -> None:
	if is_blank(derived_objective_report_id):
		raise_400('Derived objective report id is required.')

	derived_objective_report_service = get_derived_objective_report_service(principal_service)

	# noinspection DuplicatedCode
	def action() -> None:
		existing_one = derived_objective_report_service.find_tenant_and_user(derived_objective_report_id)
		if existing_one is None:
			raise_404()
		existing_tenant_id, existing_user_id = existing_one
		if existing_tenant_id != principal_service.get_tenant_id():
			raise_403()
		elif existing_user_id != principal_service.get_user_id():
			raise_403()
		# noinspection PyTypeChecker
		derived_objective_report_service.update_name(
			derived_objective_report_id, name, principal_service.get_user_id(), principal_service.get_tenant_id())

	trans(derived_objective_report_service, action)


@router.get('/indicator/derived-objective-report/delete', tags=[UserRole.CONSOLE, UserRole.ADMIN],
            response_class=Response)
async def delete_derived_objective_report_by_id(
		derived_objective_report_id: Optional[DerivedObjectiveReportId],
		principal_service: PrincipalService = Depends(get_console_principal)
) -> None:
	if is_blank(derived_objective_report_id):
		raise_400('Derived objective report id is required.')

	derived_objective_report_service = get_derived_objective_report_service(principal_service)

	# noinspection DuplicatedCode
	def action() -> None:
		# noinspection PyTypeChecker
		existing_derived_objective_report: Optional[DerivedObjectiveReport] = \
			derived_objective_report_service.find_by_id(derived_objective_report_id)
		if existing_derived_objective_report is None:
			raise_404()
		if existing_derived_objective_report.tenantId != principal_service.get_tenant_id():
			raise_403()
		if not principal_service.is_tenant_admin() and existing_derived_objective_report.userId != principal_service.get_user_id():
			raise_403()
		derived_objective_report_service.delete(derived_objective_report_id)

	trans(derived_objective_report_service, action)



from typing import Dict, List, Optional

from fastapi import APIRouter, Depends

from watchmen_auth import PrincipalService
from watchmen_indicator_kernel.common import IndicatorKernelException
from watchmen_indicator_surface.util import trans_readonly
from watchmen_meta.admin import SpaceService, TopicService
from watchmen_meta.common import ask_meta_storage, ask_snowflake_generator
from watchmen_meta.console import ConnectedSpaceService, SubjectService
from watchmen_model.admin import Space, Topic, UserRole
from watchmen_model.common import ConnectedSpaceId, SpaceId, SubjectId, TopicId
from watchmen_model.console import ConnectedSpace, Subject
from watchmen_rest import get_console_principal
from watchmen_rest.util import raise_400, raise_403, raise_404
from watchmen_utilities import ArrayHelper, is_blank

router = APIRouter()


def get_subject_service(principal_service: PrincipalService) -> SubjectService:
	return SubjectService(ask_meta_storage(), ask_snowflake_generator(), principal_service)


def get_connected_space_service(subject_service: SubjectService) -> ConnectedSpaceService:
	return ConnectedSpaceService(
		subject_service.storage, subject_service.snowflakeGenerator, subject_service.principalService)


def get_space_service(subject_service: SubjectService) -> SpaceService:
	return SpaceService(subject_service.storage, subject_service.snowflakeGenerator, subject_service.principalService)


def get_topic_service(subject_service: SubjectService) -> TopicService:
	return TopicService(subject_service.storage, subject_service.snowflakeGenerator, subject_service.principalService)


class SubjectForIndicator(Subject):
	topics: List[Topic]


@router.get(
	'/indicator/subject/name', tags=[UserRole.CONSOLE, UserRole.ADMIN], response_model=List[SubjectForIndicator])
def find_subjects_for_indicator(
		query_name: Optional[str], principal_service: PrincipalService = Depends(get_console_principal)
) -> List[SubjectForIndicator]:
	subject_service = get_subject_service(principal_service)

	def action() -> List[SubjectForIndicator]:
		if is_blank(query_name):
			subjects = subject_service.find_by_text(None, principal_service.get_tenant_id())
		else:
			subjects = subject_service.find_by_text(query_name, principal_service.get_tenant_id())
		connected_space_service = get_connected_space_service(subject_service)
		connect_ids = ArrayHelper(subjects).map(lambda x: x.connectId).distinct().to_list()
		connected_spaces = connected_space_service.find_templates_by_ids(
			connect_ids, None, principal_service.get_tenant_id())
		connect_ids = ArrayHelper(connected_spaces).map(lambda x: x.connectId).to_list()
		connected_space_map: Dict[ConnectedSpaceId, ConnectedSpace] = ArrayHelper(connected_spaces) \
			.to_map(lambda x: x.connectId, lambda x: x)
		subjects = ArrayHelper(subjects).filter(lambda x: x.connectId in connect_ids).to_list()
		space_ids = ArrayHelper(connected_spaces).map(lambda x: x.spaceId).distinct().to_list()
		space_service = get_space_service(subject_service)
		spaces = space_service.find_by_ids(space_ids, principal_service.get_tenant_id())
		space_map: Dict[SpaceId, Space] = ArrayHelper(spaces).to_map(lambda x: x.spaceId, lambda x: x)
		topic_ids = ArrayHelper(spaces).map(lambda x: x.topicIds).flatten().distinct().to_list()
		topic_service = get_topic_service(subject_service)
		topics = topic_service.find_by_ids(topic_ids, principal_service.get_tenant_id())
		topic_map: Dict[TopicId, Topic] = ArrayHelper(topics).to_map(lambda x: x.topicId, lambda x: x)

		def link_topics(subject: Subject) -> SubjectForIndicator:
			connect_id = subject.connectId
			connected_space = connected_space_map[connect_id]
			space_id = connected_space.spaceId
			space = space_map[space_id]
			all_topic_ids = space.topicIds
			all_topics = ArrayHelper(all_topic_ids).map(lambda x: topic_map[x]).to_list()
			return SubjectForIndicator(
				**subject.to_dict(),
				topics=all_topics
			)

		return ArrayHelper(subjects).map(link_topics).to_list()

	return trans_readonly(subject_service, action)


@router.get('/indicator/subject', tags=[UserRole.CONSOLE, UserRole.ADMIN], response_model=SubjectForIndicator)
def find_subject_for_indicator(
		subject_id: Optional[SubjectId], principal_service: PrincipalService = Depends(get_console_principal)
) -> SubjectForIndicator:
	if is_blank(subject_id):
		raise_400('Subject id is required.')

	subject_service = get_subject_service(principal_service)

	def action() -> SubjectForIndicator:
		subject: Optional[Subject] = subject_service.find_by_id(subject_id)
		if subject is None:
			raise_404()
		if subject.tenantId != principal_service.get_tenant_id():
			raise_403()
		connected_space_service = get_connected_space_service(subject_service)
		connected_space: Optional[ConnectedSpace] = connected_space_service.find_by_id(subject.connectId)
		if connected_space is None:
			raise IndicatorKernelException(f'Connected space not found for subject[id={subject_id}].')
		space_service = get_space_service(subject_service)
		space: Optional[Space] = space_service.find_by_id(connected_space.spaceId)
		if space is None:
			raise IndicatorKernelException(f'Space not found for subject[id={subject_id}].')
		topic_service = get_topic_service(subject_service)
		topics = topic_service.find_by_ids(space.topicIds, principal_service.get_tenant_id())
		topic_map: Dict[TopicId, Topic] = ArrayHelper(topics).to_map(lambda x: x.topicId, lambda x: x)

		all_topic_ids = space.topicIds
		all_topics = ArrayHelper(all_topic_ids).map(lambda x: topic_map[x]).to_list()
		return SubjectForIndicator(
			**subject.to_dict(),
			topics=all_topics
		)

	return trans_readonly(subject_service, action)
